#define ENTOID_PLAYER 0
#define ENTOID_GEM 1
#define ENTOID_CALVIN 2
#define ENTOID_ENEMY 3

#define EMSG_DIE 0
#define EMSG_HURT 1

// Spark -------------------------------

bool TE_Spark(Entoid& e)
{
	e.y++;
	e.timer++;
	if(e.timer>=5*3)
		return false;
	if(e.timer<0)
		e.sprite=NULL;
	else
		e.sprite=&spr_spark[e.timer/5];
	return true;
}
void SE_Spark(Entoid& e)
{
	e.type=-1;
	e.sprite=NULL;
	e.data=NULL;
	e.timer=-rnd(20);
	e.box.x=0;
	e.box.y=0;
	e.box.w=7;
	e.box.h=7;
	e.TickFunc=TE_Spark;
}

// Gem -------------------------------

bool TE_Gem(Entoid& e)
{
	if(EntoidListen(e, EMSG_DIE))
		return false;
	e.timer++;
	if(e.timer>=210)
		e.timer=rnd(100);
	if(e.timer>=200)
		e.sprite=&spr_gem[(e.timer-200)/2];
	else
		e.sprite=&spr_gem[0];
	return true;
}
void SE_Gem(Entoid& e)
{
	e.type=ENTOID_GEM;
	e.sprite=&spr_gem[0];
	e.data=NULL;
	e.timer=rnd(200);
	e.box.x=0;
	e.box.y=0;
	e.box.w=8;
	e.box.h=8;
	e.TickFunc=TE_Gem;
}

// Player -------------------------------

int anim_pwalk[9]={1,1,0,2,2,2,3,3,0};
int anim_pland[9]={3,3,3,4,0};

struct PlayerData
{
	DynVec v;
	int jump;
	bool onground;
	int doublejump;
	int* anim;
	int animlength;
	bool pspawn;
};
bool TE_Player(Entoid& e)
{
	PlayerData& pd=*((PlayerData*)e.data);

	// interactions

	if(EntoidListen(e, EMSG_DIE))
		return false;

	Entoid* target=NULL;
	if(EntoidHit(e.cbox, ENTOID_CALVIN, target))
	{
		EntoidMessage(*target, EMSG_HURT);
		if(global_score==100)
			global_win++;
		else
			global_lose++;
	}
	if(EntoidHit(e.cbox, ENTOID_ENEMY, target))
		global_lose++;
	if(e.y>200)
		global_lose++;
	Entoid* targets[100];
	int numhit=0;
	if(EntoidsHit(e.cbox, ENTOID_GEM, targets, numhit))
	{
		for(int i=0;i<numhit;i++)
		{
			Entoid& te=*targets[i];
			EntoidMessage(te, EMSG_DIE);
			global_score+=1;
			if(global_score>100) global_score=100;
			for(int si=0;si<3;si++)
				AddEntoid(SE_Spark, te.x-7+rnd(14), te.y-7+rnd(14));
		}
		playmu_playsound(sfx_gem, 50);
	}

	// movement

	DynVec& v=pd.v;

	bool moving=false;
	if(GetAsyncKeyState(VK_RIGHT)&KF_UP)
	{
		v.dx+=128;
		moving=true;
		e.flipped=false;
	}
	if(GetAsyncKeyState(VK_LEFT)&KF_UP)
	{
		v.dx-=128;
		moving=true;
		e.flipped=true;
	}
	if(!moving)
		Decay(v.dx, 64);
	if(GetAsyncKeyState('Z')&KF_UP)
	{
		if(pd.jump==0 && pd.onground)
		{
			pd.jump=1;
			playmu_playsound(sfx_jump, 50);
		}
		if(pd.jump>0 && pd.jump<20)
		{
			v.dy=-FIX(2);
			v.fy-=FIX(1);
		}
		else if(pd.jump==0 && pd.doublejump==1)
		{
			v.dy=-FIX(3);
			pd.jump=20;
			pd.doublejump=2;
		}
	}
	else
	{
		if(pd.jump>20)
			pd.jump=0;
		if(pd.jump>0 && pd.jump<20)
			pd.jump=20;
	}
	if(pd.jump>0)
		pd.jump++;
	v.dy+=24;
	
	if(GetAsyncKeyState('X')&KF_UP)
	{
		if(!pd.pspawn)
		{
			playmu_playsound(sfx_build, 30);
			pd.pspawn=true;
			int pi=rnd(numptypes-1);
			Sprite& sp=spr_platform[pi];
			int pfx=e.x+e.box.w/2-sp.width/2;
			if(moving)
			{
				if(e.flipped)
					pfx=e.x-sp.width+1;
				else
					pfx=e.x+e.box.w-1;
			}
			AddPlatform(sp, pfx, e.y+e.box.h);
			for(int spx=0;spx<sp.width;spx+=5)
				AddEntoid(SE_Spark, pfx+spx-7+rnd(14), e.y+e.box.h-7+rnd(14));
		}
	}
	else
		pd.pspawn=false;

	if(PlatformCollide(v, e.box))
	{
		if(!pd.onground) // landed from jump
		{
			e.timer=0;
			pd.anim=anim_pland;
			pd.animlength=5;
		}
		pd.onground=true;
		if(pd.doublejump==2)
			pd.doublejump=1;
	}
	else
		pd.onground=false;

	// walk animation
	if(moving && pd.onground && e.timer==-1) e.timer=0;
	if(e.timer>-1) e.timer++;
	if(e.timer>=3*pd.animlength)
	{
		e.timer=0;
		pd.anim=anim_pwalk;
		pd.animlength=9;
	}
	if(e.timer==0 && !(moving && pd.onground)) // stop walking
		e.timer=-1;

	if(e.timer>-1 && pd.anim!=NULL)
		e.sprite=&spr_room[pd.anim[e.timer/3]];
	else
		e.sprite=&spr_room[0];
	
	if(!pd.onground) // in air
		e.sprite=&spr_room[5];

	Limit(v.dx, -FIX(2), FIX(2));
	Limit(v.dy, -FIX(3), FIX(4));
	v.fx+=v.dx;
	v.fy+=v.dy;
	Limit(v.fx, -FIX(320*2), FIX(320*3-12));
	e.x=INT(v.fx);
	e.y=INT(v.fy);

	return true;
}
void SE_Player(Entoid& e)
{
	e.type=ENTOID_PLAYER;
	e.sprite=&spr_room[0];
	e.layer=2; // top layer
	e.data=new PlayerData;
	GCAdd(e.data, GCNEW);
	e.box.x=9;
	e.box.y=10;
	e.box.w=12;
	e.box.h=20;
	// center on bottom middle
	e.x-=e.box.w/2;
	e.y-=e.box.h;

	e.timer=-1;

	PlayerData& pd=*((PlayerData*)e.data);
	pd.v.fx=FIX(e.x);
	pd.v.fy=FIX(e.y);
	pd.v.dx=0;
	pd.v.dy=0;
	pd.jump=0;
	pd.onground=false;
	pd.doublejump=0;
	pd.anim=NULL;
	pd.animlength=0;
	pd.pspawn=false;

	e.TickFunc=TE_Player;
}

// Calvin -------------------------------

struct CalvinData
{
	int frame;
	int mphase;
	int sy;
	int hurt;
};
bool TE_Calvin(Entoid& e)
{
	CalvinData& cd=*((CalvinData*)e.data);

	if(EntoidListen(e, EMSG_DIE))
		return false;
	if(EntoidListen(e, EMSG_HURT))
		cd.hurt++;
	else
		cd.hurt=0;
	
	cd.mphase++;
	if(cd.sy>120)
		cd.sy--;
	e.y=cd.sy+(int)(sin((float)cd.mphase*0.03f)*40);
	
	e.timer++;
	if(e.timer>=200)
	{
		e.timer=rnd(150);
		cd.frame=1-cd.frame;
	}
	e.sprite=&spr_calvin[cd.frame];
	
/*	if(entoids[0].x+6>e.x+31)
		e.flipped=true;
	else
		e.flipped=false;
*/	
	if(cd.hurt>0)
		e.sprite=&spr_calvin[1+((cd.hurt/2)&1)];
	return true;
}
void SE_Calvin(Entoid& e)
{
	e.type=ENTOID_CALVIN;
	e.sprite=&spr_calvin[0];
	e.layer=1; // middle layer
	e.timer=0;
	e.box.x=10;
	e.box.y=17;
	e.box.w=62;
	e.box.h=31;
	
	e.data=new CalvinData;
	GCAdd(e.data, GCNEW);
	CalvinData& cd=*((CalvinData*)e.data);
	cd.frame=0;
	cd.mphase=0;
	cd.sy=e.y;
	cd.hurt=0;
	
	e.TickFunc=TE_Calvin;
}

// Drone -------------------------------

struct DroneData
{
	int deposit;
	int dy;
	int fy;
};
bool TE_Drone(Entoid& e)
{
	DroneData& dd=*((DroneData*)e.data);

	e.timer++;
	if(e.timer==50)
		playmu_playsound(sfx_drone, 30);
	if(e.y<-240*4-50)
	{
		e.timer=-rnd(150);
		e.y=240;
		dd.fy=FIX(240);
		dd.dy=0;
		dd.deposit=0;
		e.x=-320*2+50+rnd(320*5-100);
	}

	if(dd.dy==-FIX(4))
	{
		if(dd.deposit==0 && rnd(30)==0)
			dd.deposit=1;
		if(dd.deposit==1 && rnd(50)==0)
			dd.deposit=2;
	}
	
	if(e.timer>0)
	{
		dd.dy-=4;
		if(dd.dy<-FIX(4))
			dd.dy=-FIX(4);
		dd.fy+=dd.dy;
		e.y=INT(dd.fy);
	}

	// deposit gem	
	if(dd.deposit==1 && (e.timer%6)==0)
		AddEntoid(SE_Gem, e.x+4, e.y+30);
	
	if(e.timer>=0)
		e.sprite=&spr_drone[(e.timer/2)&1];
	return true;
}
void SE_Drone(Entoid& e)
{
	e.sprite=&spr_drone[0];
	e.layer=1; // middle layer
	e.timer=-100-rnd(150);
	e.box.x=0;
	e.box.y=0;
	e.box.w=16;
	e.box.h=52;

	e.data=new DroneData;
	GCAdd(e.data, GCNEW);
	DroneData& dd=*((DroneData*)e.data);
	dd.dy=0;
	dd.fy=FIX(240);
	dd.deposit=0;
	
	e.TickFunc=TE_Drone;
}

// Enemy -------------------------------

struct EnemyData
{
	DynVec v;
	bool splash;
	bool onground;
};
bool TE_Enemy(Entoid& e)
{
	EnemyData& ed=*((EnemyData*)e.data);

	e.timer++;

	ed.v.dy+=8;
	if(ed.v.dy>FIX(4))
		ed.v.dy=FIX(4);
	ed.v.fy+=ed.v.dy;

	if(PlatformCollide(ed.v, e.box))
	{
		if(!ed.onground)
		{
			e.timer=0;
			playmu_playsound(sfx_splat, 30);
		}
		ed.onground=true;
	}
	
	if(!ed.splash && e.y<200)
	{
		playmu_playsound(sfx_splash, 10);
		ed.splash=true;
		for(int si=0;si<6;si++)
			AddEntoid(SE_Spark, e.x-3+rnd(14), e.y-25+rnd(14));
	}

	if(ed.v.dy<0)
		e.sprite=&spr_enemy[0];
	else
		e.sprite=&spr_enemy[2];
	if(abs(ed.v.dy)<32)
		e.sprite=&spr_enemy[1];

	if(ed.onground)
	{
		if(e.timer<3)
			e.sprite=&spr_enemy[3];
		else
			e.sprite=&spr_enemy[4];
	}
	
	if(entoids[0].x+6<e.x+10)
		e.flipped=true;
	else
		e.flipped=false;

	e.y=INT(ed.v.fy);
	if(e.y>240)
		return false;
	return true;
}
void SE_Enemy(Entoid& e)
{
	e.type=ENTOID_ENEMY;
	e.sprite=&spr_enemy[0];
	e.layer=1; // middle layer
	e.box.x=6;
	e.box.y=7;
	e.box.w=8;
	e.box.h=10;

	e.data=new EnemyData;
	GCAdd(e.data, GCNEW);
	EnemyData& ed=*((EnemyData*)e.data);
	ed.v.fx=FIX(e.x);
	ed.v.fy=FIX(240);
	ed.v.dx=0;
	ed.v.dy=-FIX(3)-rnd(FIX(3));
	ed.splash=false;
	ed.onground=false;
	
	e.TickFunc=TE_Enemy;
}

